""" spec2nii - tool for conversion of various MRS data formats to NIFTI format.

This module contains the main class to be called as a script (through the main function).

Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""

import argparse
import sys
import os.path as op
from pathlib import Path
import json
from nibabel.nifti2 import Nifti2Image
from spec2nii import __version__ as spec2nii_ver
from numpy import isclose
# There are case specific imports below


class Spec2niiError(Exception):
    pass


class spec2nii:
    def __init__(self):
        cite_str = "Clarke WT, Bell TK, Emir UE, Mikkelsen M, Oeltzschner G, Shamaei A, Soher BJ, Wilson M. "\
                   "NIfTI-MRS: A standard data format for magnetic resonance spectroscopy. "\
                   "Magn Reson Med. 2022. doi: 10.1002/mrm.29418."
        parser = argparse.ArgumentParser(
            description='Convert raw spectroscopy data to NIfTI format.',
            epilog=f"If you use spec2nii please cite: {cite_str}")
        parser.add_argument('-v', '--version', action='version', version=spec2nii_ver)

        subparsers = parser.add_subparsers(title='spec2nii subcommands')

        def add_common_parameters(subparser):
            # Add options that are common to all subcommands
            subparser.add_argument('-j', '--json', help='Create json sidecar.', action='store_true')
            subparser.add_argument("-f", "--fileout", type=str,
                                   help="Output file base name (default = input file name)")
            subparser.add_argument("-o", "--outdir", type=Path,
                                   help="Output location (default = .)", default='.')
            subparser.add_argument('--nifti1', action='store_true')
            subparser.add_argument("--override_nucleus", type=str, nargs='+',
                                   help="Override ResonantNucleus field with input(s). E.g. '2H'.")
            subparser.add_argument("--override_frequency", type=float, nargs='+',
                                   help="Override SpectrometerFrequency field with input(s). Input in MHz.")
            subparser.add_argument("--override_dwelltime", type=float,
                                   help="Override dwell time field with input. Input in seconds.")
            subparser.add_argument(
                '--anon',
                action='store_true',
                help="Create file without sensitive metadata. For greater control use spec2nii anon.")
            subparser.add_argument('--verbose', action='store_true')

            return subparser

        # Auto subcommand - heuristic ID of file type
        parser_auto = subparsers.add_parser('auto', help='Attempt automatic identification and conversion.')
        parser_auto.add_argument('file', help='file to convert', type=Path)
        parser_auto = add_common_parameters(parser_auto)
        parser_auto.set_defaults(func=self.auto)

        # Handle twix subcommand
        parser_twix = subparsers.add_parser('twix', help='Convert from Siemens .dat twix format.')
        parser_twix.add_argument('file', help='file to convert', type=Path)
        group = parser_twix.add_mutually_exclusive_group(required=True)
        group.add_argument("-v", "--view", help="View contents of twix file, no files converted", action='store_true')
        group.add_argument('-e', '--evalinfo', type=str, help='evalInfo flag to convert')
        parser_twix.add_argument("-m", "--multiraid", type=int, help="Select multiraid file to load"
                                 " (default = 2 i.e. 2nd file)", default=2)
        parser_twix.add_argument("-q", "--quiet", help="Suppress text output", action='store_true')
        for idx in range(5, 8):
            parser_twix.add_argument(f"-d{idx}", f"--dim{idx}", type=str, help=f"Specify dim {idx} loop counter.")
            parser_twix.add_argument(f"-t{idx}", f"--tag{idx}", type=str, help=f"Specify dim {idx} NIfTI MRS tag.")
        parser_twix.add_argument('--remove_os', action='store_true', help='Remove time-domain oversampling.')
        parser_twix = add_common_parameters(parser_twix)
        parser_twix.set_defaults(func=self.twix)

        # Handle dicom subcommand
        parser_dicom = subparsers.add_parser('dicom', help='Convert from Siemens DICOM format.')
        parser_dicom.add_argument('file', help='file or directory to convert', type=str)
        parser_dicom.add_argument(
            "-t",
            "--tag",
            type=str,
            default='DIM_DYN',
            help="Specify NIfTI MRS tag used for 5th dimension if multiple files are passed. "
                 "Defaults to DIM_DYN.")
        parser_dicom.add_argument('--voi', action='store_true', help='Output VOI as single voxel NIfTI mask.')
        parser_dicom = add_common_parameters(parser_dicom)
        parser_dicom.set_defaults(func=self.dicom)

        # Handle rda subcommand
        parser_rda = subparsers.add_parser('rda', help='Convert from Siemens spectroscopy .rda format.')
        parser_rda.add_argument('file', help='file to convert', type=Path)
        parser_rda = add_common_parameters(parser_rda)
        parser_rda.set_defaults(func=self.rda)

        # Handle UIH DICOM subcommand
        parser_uih_dicom = subparsers.add_parser('uih', help='Convert from UIH DICOM format.')
        parser_uih_dicom.add_argument('file', help='file or directory to convert', type=str)
        parser_uih_dicom.add_argument("-t", "--tag", type=str, help="Specify NIfTI MRS tag used for 5th "
                                                                    "dimension if multiple files are passed.")
        parser_uih_dicom = add_common_parameters(parser_uih_dicom)
        parser_uih_dicom.set_defaults(func=self.uih_dicom)

        # Handle philips subcommand
        parser_philips = subparsers.add_parser('philips', help='Convert from Philips spar/sdat format.')
        parser_philips.add_argument('sdat', help='SDAT file', type=Path)
        parser_philips.add_argument('spar', help='SPAR file', type=Path)
        parser_philips.add_argument(
            "-t", "--tags",
            type=str,
            nargs='+',
            default=[None, None, None],
            help="Specify NIfTI MRS tags used for higher (5th-7th) dimensions. "
                 "Defaults to DIM_DYN if more than one spectrum is present. "
                 "Can be used to create singleton higher dimensions.")
        parser_philips.add_argument(
            "-s", "--shape",
            type=int,
            nargs='+',
            default=None,
            help="Specify the shape of higher (5th-7th) dimensions. Applies numpy style reshaping.")
        parser_philips.add_argument(
            "--special",
            type=str,
            default=None,
            help="Identify special case sequence. Options: 'hyper', 'hyper-ref'.")
        parser_philips = add_common_parameters(parser_philips)
        parser_philips.set_defaults(func=self.philips)

        # Handle philips data/list subcommand
        parser_p_dl = subparsers.add_parser('philips_dl', help='Convert from Philips data/list format.')
        parser_p_dl.add_argument('data', help='.data file', type=Path)
        parser_p_dl.add_argument('list', help='.list file', type=Path)
        parser_p_dl.add_argument('aux', help='Auxiliary file: .SPAR or DICOM file', type=Path)
        # for idx in range(5, 8):
        #     parser_p_dl.add_argument(f"-d{idx}", f"--dim{idx}", type=str, help=f"Specify dim {idx} loop counter.")
        #     parser_p_dl.add_argument(f"-t{idx}", f"--tag{idx}", type=str, help=f"Specify dim {idx} NIfTI MRS tag.")
        parser_p_dl.add_argument(
            "--special",
            type=str,
            default=None,
            help="Identify special case sequence. Options: 'hyper'.")
        parser_p_dl = add_common_parameters(parser_p_dl)
        parser_p_dl.set_defaults(func=self.philips_dl)

        # Handle philips DICOM subcommand
        parser_philips_dicom = subparsers.add_parser('philips_dcm', help='Convert from Philips DICOM format.')
        parser_philips_dicom.add_argument('file', help='file or directory to convert', type=str)
        parser_philips_dicom.add_argument(
            "-t",
            "--tag",
            type=str,
            default='DIM_DYN',
            help="Specify NIfTI MRS tag used for 5th dimension if multiple files are passed. "
                 "Defaults to DIM_DYN.")
        parser_philips_dicom = add_common_parameters(parser_philips_dicom)
        parser_philips_dicom.set_defaults(func=self.philips_dicom)

        # Handle GE subcommand
        parser_ge = subparsers.add_parser('ge', help='Convert from GE p-file format.')
        parser_ge.add_argument('file', help='file to convert', type=Path)
        parser_ge = add_common_parameters(parser_ge)
        parser_ge.set_defaults(func=self.ge)

        # # Handle ismrmrd subcommand
        # parser_ismrmrd = subparsers.add_parser('ismrmrd', help='Convert from ismrmrd format.')
        # parser_ismrmrd.add_argument('file', help='file to convert', type=str)
        # parser_ismrmrd.add_argument('a', type=str, help='placeholder')
        # parser_ismrmrd.set_defaults(func=self.ismrmrd)

        # Handle text subcommand
        parser_txt = subparsers.add_parser('text', help='Convert from plain text format.')
        parser_txt.add_argument('file', help='file to convert', type=str)
        parser_txt.add_argument("-i", "--imagingfreq", type=float,
                                help="Imaging (central) frequency in MHz", required=True)
        parser_txt.add_argument("-b", "--bandwidth", type=float,
                                help="Receiver bandwidth (sweepwidth) in Hz.", required=True)
        parser_txt.add_argument("-n", "--nucleus", type=str,
                                help="Nucleus string. e.g. 1H or 31P.", required=True)
        parser_txt.add_argument("-a", "--affine", type=str, help="NIfTI affine file", required=False, metavar='<file>')
        parser_txt = add_common_parameters(parser_txt)
        parser_txt.set_defaults(func=self.text)

        # jMRUI formats (.txt and .mrui)
        parser_jmrui = subparsers.add_parser('jmrui', help='Convert from jMRUI text or binary formats.')
        parser_jmrui.add_argument('file', help='file to convert', type=Path)
        parser_jmrui.add_argument("-a", "--affine", type=str, help="NIfTI affine file", required=False,
                                  metavar='<file>')
        parser_jmrui = add_common_parameters(parser_jmrui)
        parser_jmrui.set_defaults(func=self.jmrui)

        # LCModel RAW/H2O formats
        parser_raw = subparsers.add_parser('raw', help='Convert from LCModel RAW text format.')
        parser_raw.add_argument('file', help='file to convert', type=str)
        parser_raw.add_argument("-n", "--nucleus", type=str,
                                help="Nucleus string. e.g. 1H or 31P.", required=True)
        parser_raw.add_argument("-i", "--imagingfreq", type=float,
                                help="Optional. Imaging (central) frequency in MHz", required=False)
        parser_raw.add_argument("-b", "--bandwidth", type=float,
                                help="Optional. Receiver bandwidth (spectral width) in Hz.", required=False)
        parser_raw.add_argument("-a", "--affine", type=str, help="NIfTI affine file", required=False, metavar='<file>')
        parser_raw = add_common_parameters(parser_raw)
        parser_raw.set_defaults(func=self.raw)

        # Bruker format
        parser_bruker = subparsers.add_parser('bruker', help='Convert from Bruker data format.')
        parser_bruker.add_argument('file', help='2dseq file to convert', type=str)
        parser_bruker.add_argument('-q', '--query', action='append', default=[])
        parser_bruker.add_argument('-m', '--mode', type=str, default='2DSEQ', choices=['2DSEQ', 'FID'])
        parser_bruker.add_argument('-d', '--dump_headers',
                                   help='Dump bruker header files into json header extension',
                                   action='store_true')
        parser_bruker = add_common_parameters(parser_bruker)
        parser_bruker.set_defaults(func=self.bruker)

        # Varian format

        parser_varian = subparsers.add_parser('varian', help='Convert from the Varian/OpenVnmrJ data format')
        parser_varian.add_argument(
            'file', help='Path to the varian .fid directory, containing procpar and fid files', type=str)
        parser_varian.add_argument("-t", "--tag6", type=str, default='DIM_DYN',
                                   help="Specify 6th dimension NIfTI MRS tag (default DIM_DYN).")
        parser_varian.add_argument('-d', '--dump_headers',
                                   help='Dump varian header information into json header extension',
                                   action='store_true')
        parser_varian = add_common_parameters(parser_varian)
        parser_varian.set_defaults(func=self.varian)

        # Additional functions - anonymise and dump
        parser_anon = subparsers.add_parser('anon', help='Anonymise existing NIfTI-MRS file.')
        parser_anon.add_argument('file', help='file to anonymise', type=Path)
        parser_anon.add_argument("-r", "--remove", type=str, metavar='KEY',
                                 action='append',
                                 help="Explicitly remove key. Argument may be repeated")
        parser_anon.add_argument("-f", "--fileout", type=str,
                                 help="Output file base name (default = input file name)")
        parser_anon.add_argument("-o", "--outdir", type=Path,
                                 help="Output location (default = .)", default='.')
        parser_anon.add_argument('-v', '--verbose', action='store_true')
        parser_anon.set_defaults(func=self.anon,
                                 json=False,
                                 nifti1=False,
                                 override_nucleus=None,
                                 override_frequency=None,
                                 override_dwelltime=None,
                                 anon=False)

        parser_dump = subparsers.add_parser('dump', help='Dump contents of headers from existing NIfTI-MRS file.')
        parser_dump.add_argument('file', help='NIfTI-MRS file', type=Path)
        parser_dump.set_defaults(func=self.dump,
                                 outdir=None,
                                 nifti1=False)

        parser_extract = subparsers.add_parser('extract', help='Generate json sidecar from existing NIfTI-MRS file.')
        parser_extract.add_argument('file', help='NIFTI-MRS file', type=Path)
        parser_extract.add_argument("-o", "--outdir", type=Path,
                                    help="Output location (default: directory of input file)", default=None)
        parser_extract.set_defaults(func=self.extract,
                                    nifti1=False)

        parser_insert = subparsers.add_parser('insert', help='Insert json formatted file into existing NIfTI-MRS file.')
        parser_insert.add_argument('file', help='NIFTI-MRS file', type=Path)
        parser_insert.add_argument('json_file', help='JSON file to insert', type=Path)
        parser_insert.add_argument("--dwelltime", type=float,
                                   help="Specify a new dwelltime (1/bandwidth, pixdim[4]) value in seconds.")
        parser_insert.add_argument("-f", "--fileout", type=str,
                                   help="Output file base name (default = input file name)")
        parser_insert.add_argument("-o", "--outdir", type=Path,
                                   help="Output location (default = .)", default='.')
        parser_insert.set_defaults(func=self.insert,
                                   nifti1=False,
                                   json=False,
                                   override_nucleus=None,
                                   override_frequency=None,
                                   override_dwelltime=None,
                                   verbose=False,
                                   anon=False)

        if len(sys.argv) == 1:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        args = parser.parse_args()

        self.fileoutNames = []
        self.imageOut = []
        self.orientationInfoOut = []
        self.dwellTimes = []
        self.metaData = []

        self.outputDir = args.outdir

        args.func(args)

        if self.imageOut:
            self.implement_overrides(args)

            self.insert_spectralwidth()

            if args.anon:
                from spec2nii.anonymise import anon_nifti_mrs
                for idx, nifti_mrs_img in enumerate(self.imageOut):
                    self.imageOut[idx] = anon_nifti_mrs(nifti_mrs_img, verbose=args.verbose)

            self.validate_output()
            self.write_output(args.json, args.nifti1)
            self.validate_write(args.verbose)
            if args.verbose:
                print(f'Please cite {cite_str}.')
        elif hasattr(args, 'view') and not args.view:
            print('No files to write.')

    def implement_overrides(self, args):
        """Implement any command line overrides for essential parameters."""
        for nifti_mrs_img in self.imageOut:
            if args.override_dwelltime:
                nifti_mrs_img.set_dwell_time(args.override_dwelltime)

            if args.override_nucleus or args.override_frequency:
                from nibabel.nifti1 import Nifti1Extension
                hdr_ext_codes = nifti_mrs_img.header.extensions.get_codes()
                index = hdr_ext_codes.index(44)
                original = json.loads(nifti_mrs_img.header.extensions[index].get_content())

                if args.override_nucleus:
                    original['ResonantNucleus'] = args.override_nucleus
                if args.override_frequency:
                    original['SpectrometerFrequency'] = args.override_frequency
                json_s = json.dumps(original)
                new_ext = Nifti1Extension(44, json_s.encode('UTF-8'))
                nifti_mrs_img.header.extensions.clear()
                nifti_mrs_img.header.extensions.append(new_ext)

    def insert_spectralwidth(self):
        """Ensure that the correct spectral width is inserted into the header extension"""
        for nifti_mrs_img in self.imageOut:
            if 'SpectralWidth' in nifti_mrs_img.hdr_ext\
                    and not isclose(
                        nifti_mrs_img.hdr_ext['SpectralWidth'],
                        1 / nifti_mrs_img.dwelltime,
                        atol=1E-2):
                nifti_mrs_img.remove_hdr_field('SpectralWidth')
                nifti_mrs_img.add_hdr_field('SpectralWidth', 1 / nifti_mrs_img.dwelltime)
            else:
                nifti_mrs_img.add_hdr_field('SpectralWidth', 1 / nifti_mrs_img.dwelltime)

    def validate_output(self):
        """Run NIfTI MRS validation on output."""
        import nifti_mrs.validator as validate
        # Currently this repeats the validation before the save.
        # But useful here to do exception handling
        for f_out, nifti_mrs_img in zip(self.fileoutNames, self.imageOut):
            try:
                validate.validate_nifti_mrs(nifti_mrs_img)
            except (
                    validate.headerExtensionError,
                    validate.niftiDataError,
                    validate.niftiHeaderError) as exc:

                raise Spec2niiError(f'Generated file {f_out} failed validation.') from exc

    def write_output(self, write_json=False, nifti1=False):
        """Write any NIfTI MRS objects stored.
        If write_json is true also write meta-data as sidecar.
        """
        self.outputDir.mkdir(parents=True, exist_ok=True)

        for f_out, nifti_mrs_img in zip(self.fileoutNames, self.imageOut):
            out = self.outputDir / (f_out + '.nii.gz')

            if nifti1:
                # If nifti1 is requested just remake the file.
                # This is more than a bit hacky, but avoids passing the option to every end-point.
                from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
                gen_nifti_mrs_hdr_ext(
                    nifti_mrs_img[:],
                    nifti_mrs_img.dwelltime,
                    nifti_mrs_img.hdr_ext,
                    nifti_mrs_img.getAffine('voxel', 'world'),
                    nifti_version=1)\
                    .save(out)
            else:
                nifti_mrs_img.save(out)

            if write_json:
                out_json = self.outputDir / (f_out + '.json')
                with open(out_json, 'w') as fp:
                    json.dump(json.loads(nifti_mrs_img.header.extensions[0].get_content()), fp, indent=4)

    class NIfTIMRSWriteError(IOError):
        pass

    def validate_write(self, verbose):
        for f_out in self.fileoutNames:
            out = self.outputDir / (f_out + '.nii.gz')
            if out.exists() and verbose:
                print(f'Output {out.name} written to {out.parent}')
            elif out.exists():
                pass
            else:
                raise self.NIfTIMRSWriteError(f'Output {out.name} in {out.parent} not found!')

    # Attempt automatic file type identification
    def auto(self, args):
        """Attempt automatic file type identification and conversion

        Currently handles: Twix, RDA, SPAR/SDAT, GE pfile, DICOM
        """
        from warnings import warn
        warn('Automatic conversion is an experimental feature. Please verify the output carefully.')

        # Twix format - ID by extension
        if args.file.suffix.lower() == '.dat':
            print("Attempting conversion as Siemens Twix (.dat), evalinfo = 'image'")
            setattr(args, 'evalinfo', 'image')
            setattr(args, 'quiet', False)
            setattr(args, 'view', False)
            for idx in range(5, 8):
                setattr(args, f'dim{idx}', None)
                setattr(args, f'tag{idx}', None)
            setattr(args, 'remove_os', False)
            self.twix(args)

        # Siemens RDA format - ID by extension
        elif args.file.suffix.lower() == '.rda':
            print('Attempting conversion as Siemens RDA (.rda)')
            self.rda(args)

        # Philips SPAR/SDAT format - ID by extension(s)
        elif args.file.suffix.lower() in ('.spar', '.sdat'):
            print('Attempting conversion as Philips SPAR/SDAT pair')
            if args.file.with_suffix('.SPAR').is_file()\
                    and args.file.with_suffix('.SDAT').is_file():
                setattr(args, 'spar', args.file.with_suffix('.SPAR'))
                setattr(args, 'sdat', args.file.with_suffix('.SDAT'))
            else:
                raise Spec2niiError(
                    'Unable to find both files in SPAR/SDAT pair.'
                    'Please use manual selection of subcomands. '
                    'e.g. spec2nii philips ...')
            setattr(args, 'tags', ["DIM_DYN", None, None])
            setattr(args, 'shape', None)
            setattr(args, 'special', None)
            self.philips(args)

        # Philips DATA/LIST - Reject because of unknown aux file format.
        elif args.file.suffix.lower() in ('.data', '.list'):
            raise Spec2niiError(
                'Automatic conversion not setup for data/list conversion. '
                'Please use manual selection of subcomand. '
                'e.g. spec2nii philips_dl ...')

        # GE pfile (.7) format - ID by extension
        elif args.file.suffix.lower() == '.7':
            self.ge(args)

        # If no matches assume DICOM - ID by loading file
        else:
            import pydicom as pdcm
            try:
                if args.file.is_dir():
                    files_in = \
                        sorted(args.file.rglob('*.IMA')) + \
                        sorted(args.file.rglob('*.ima')) + \
                        sorted(args.file.rglob('*.dcm'))
                    file = pdcm.dcmread(files_in[0])
                else:
                    file = pdcm.dcmread(args.file)

                manufacturer = file.Manufacturer
                setattr(args, 'tag', None)
                if manufacturer.lower() == 'siemens':
                    print('Attempting conversion as Siemens DICOM.')
                    setattr(args, 'voi', False)
                    self.dicom(args)
                elif manufacturer.lower() == 'uih':
                    print('Attempting conversion as UIH DICOM.')
                    self.uih_dicom(args)
                elif manufacturer.lower() == 'philips medical systems':
                    print('Attempting conversion as Philips DICOM.')
                    self.philips_dicom(args)
                else:
                    raise Spec2niiError(f'Unknown DICOM manufacturer {manufacturer}.')
            # No successful ID as DICOM - fail at automatic load.
            except pdcm.errors.InvalidDicomError:
                raise Spec2niiError(
                    'Unable to automatically identify file type. '
                    'Please use manual selection of subcomands. '
                    'e.g. spec2nii twix ..., spec2nii philips ...')

    # Start of the specific format handling functions.
    # Siemens twix (.dat) format
    def twix(self, args):
        """ Twix format handler."""
        from mapvbvd import mapVBVD
        from spec2nii.Siemens.twixfunctions import process_twix, examineTwix

        # Call mapvbvd to load the twix file.
        twixObj = mapVBVD(args.file, quiet=args.quiet)

        if args.view:
            examineTwix(twixObj, op.basename(args.file), args.multiraid)
        else:
            if isinstance(twixObj, list):
                twixObj = twixObj[args.multiraid - 1]

            if not args.quiet:
                print(f"Converting twix file {args.file}.")
                print(f'Looking for evalinfo flag {args.evalinfo}.')

            overrides = {'dims': (args.dim5, args.dim6, args.dim7),
                         'tags': (args.tag5, args.tag6, args.tag7)}

            self.imageOut, self.fileoutNames = process_twix(twixObj,
                                                            args.fileout,
                                                            args.file.name,
                                                            args.evalinfo,
                                                            overrides,
                                                            args.quiet,
                                                            args.verbose,
                                                            remove_os=args.remove_os)

    # (Siemens) DICOM (.ima) format
    def dicom(self, args):
        """Siemens DICOM format handler."""
        from spec2nii.Siemens.dicomfunctions import multi_file_dicom
        path_in = Path(args.file)
        if path_in.is_dir():
            # Look for typical dicom file extensions
            files_in = sorted(path_in.rglob('*.IMA')) + \
                sorted(path_in.rglob('*.ima')) + \
                sorted(path_in.rglob('*.dcm'))

            # If none found look for all files
            if len(files_in) == 0:
                files_in = sorted([x for x in path_in.iterdir() if x.is_file()])

            print(f'Found {len(files_in)} files.')
        else:
            print('Single file conversion.')
            files_in = [path_in]

        self.imageOut, self.fileoutNames = multi_file_dicom(files_in, args.fileout, args.tag, args.verbose)

        # VOI mask - TO DO extend to other formats
        if args.voi:
            import numpy as np
            from nibabel import save
            for img, f_out in zip(self.imageOut, self.fileoutNames):
                out = self.outputDir / (f_out + '_voi.nii.gz')
                mask_data = np.ones((1, 1, 1), int)
                save(
                    Nifti2Image(
                        mask_data,
                        img.hdr_ext['VOI'],
                        dtype=mask_data.dtype),
                    out)

    # Siemens RDA (.rda) format
    def rda(self, args):
        """Siemens RDA format handler."""
        from spec2nii.Siemens.rda import convert_rda
        self.imageOut, self.fileoutNames = convert_rda(args.file, args.fileout, args.verbose)

    # (UIH) DICOM (.dcm) format
    def uih_dicom(self, args):
        """UIH DICOM format handler."""
        from spec2nii.uih import multi_file_dicom
        path_in = Path(args.file)
        if path_in.is_dir():
            # Look for typical dicom file extensions
            files_in = sorted(path_in.glob('*.dcm'))

            # If none found look for all files
            if len(files_in) == 0:
                files_in = sorted([x for x in path_in.iterdir() if x.is_file()])

            print(f'Found {len(files_in)} files.')
        else:
            print('Single file conversion.')
            files_in = [path_in]

        self.imageOut, self.fileoutNames = multi_file_dicom(files_in, args.fileout, args.tag, args.verbose)

    # Philips SDAP/SPAR handler
    def philips(self, args):
        # philips specific imports
        from spec2nii.Philips.philips import read_sdat_spar_pair

        self.imageOut, self.fileoutNames = read_sdat_spar_pair(
            args.sdat,
            args.spar,
            args.shape,
            args.tags,
            args.fileout,
            args.special)

    # Philips data/list (+SPAR) handler
    def philips_dl(self, args):
        # philips specific imports
        from spec2nii.Philips.philips_data_list import read_data_list_pair

        # TO DO
        # overrides = {'dims': (args.dim5, args.dim6, args.dim7),
        #              'tags': (args.tag5, args.tag6, args.tag7)}

        # self.imageOut, file_names = read_data_list_pair(args.data, args.list, args.spar, overrides)
        self.imageOut, file_names = read_data_list_pair(args.data, args.list, args.aux, args.special)

        # name of output
        if args.fileout:
            mainStr = args.fileout
        else:
            mainStr = args.data.stem

        for fn in file_names:
            self.fileoutNames.append(mainStr + '_' + fn)

    # Philips DICOM handler
    def philips_dicom(self, args):
        """Philips DICOM format handler."""
        from warnings import warn
        from spec2nii.Philips.philips_dcm import multi_file_dicom
        warn('This Philips DICOM conversion routine has limited testing outside vendor PRESS and MEGA sequences.'
             ' Please get in contact with test data to help improve it.')
        path_in = Path(args.file)
        if path_in.is_dir():
            # Look for typical dicom file extensions
            files_in = sorted(path_in.glob('*.dcm'))

            # If none found look for all files
            if len(files_in) == 0:
                files_in = sorted([x for x in path_in.iterdir() if x.is_file()])

            print(f'Found {len(files_in)} files.')
        else:
            print('Single file conversion.')
            files_in = [path_in]

        self.imageOut, self.fileoutNames = multi_file_dicom(files_in, args.fileout, args.tag, args.verbose)

    # GE p file handler
    def ge(self, args):
        # ge specific imports
        from spec2nii.GE.ge_pfile import read_pfile
        self.imageOut, self.fileoutNames = read_pfile(args.file, args.fileout)

    def text(self, args):
        from spec2nii.other_formats import text
        # Simply pass through to the specific text function
        self.imageOut, self.fileoutNames = text(args)

    def jmrui(self, args):
        from spec2nii.jmrui import jmrui_format
        # Pass straight through to dedicated function
        self.imageOut, self.fileoutNames = jmrui_format(args)

    def raw(self, args):
        from spec2nii.other_formats import lcm_raw
        self.imageOut, self.fileoutNames = lcm_raw(args)

    # Bruker 2dseq files with FG_COMPLEX
    def bruker(self, args):
        from spec2nii.bruker import read_bruker

        self.imageOut, self.fileoutNames = read_bruker(args)

    # Varian parser
    def varian(self, args):
        from spec2nii.varian_importer import read_varian
        self.imageOut, self.fileoutNames = read_varian(args)

    # Anonymise function
    def anon(self, args):
        from spec2nii.anonymise import anon_file
        self.imageOut, self.fileoutNames = anon_file(args)

    # Dump function
    @staticmethod
    def dump(args):
        from spec2nii.other import dump_headers
        dump_headers(args)

    # Extract function
    @staticmethod
    def extract(args):
        from spec2nii.other import extract_hdr_ext
        extract_hdr_ext(args)

    # Insert function
    def insert(self, args):
        from spec2nii.other import insert_hdr_ext
        self.imageOut, self.fileoutNames = insert_hdr_ext(args)


def main(*args):
    spec2nii(*args)
    return 0



""" spec2nii module containing functions specific to Siemens rda format
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2022 University of Oxford
"""
import re
from datetime import datetime

import numpy as np

from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

from spec2nii.dcm2niiOrientation.orientationFuncs import dcm_to_nifti_orientation
from spec2nii import __version__ as spec2nii_ver


class MRSINotHandledError(Exception):
    pass


def _locale_float(x):
    """Handle locale specific floating point representations in header

    :param x: Header value as string with either . or , decimal separator
    :type x: str
    :return: Value cast to float
    :rtype: float
    """
    try:
        return float(x)
    except ValueError as e:
        if ',' in x:
            return float(x.replace(',', '.'))
        else:
            raise e


def convert_rda(rda_path, fname_out, verbose):
    hdr_st = re.compile(r'>>> Begin of header <<<')
    hdr_val = re.compile(r'^([\d\w\[\],]+): (.*)\r\n')
    hdr_end = re.compile(r'>>> End of header <<<')
    hdr = {}

    with open(rda_path, 'rb') as fp:
        for line in fp:
            try:
                if hdr_st.search(line.decode()):
                    pass
                    # print('header found')
                elif hdr_end.search(line.decode()):
                    # print('header end')
                    break
                else:
                    match = hdr_val.search(line.decode())
                    if len(match.groups()) < 2:
                        hdr[match[1]] = None
                    else:
                        hdr[match[1]] = match[2]
            except UnicodeDecodeError:
                print('Trying latin-1 encoding.')
                if hdr_st.search(line.decode('latin-1')):
                    pass
                    # print('header found')
                elif hdr_end.search(line.decode('latin-1')):
                    # print('header end')
                    break
                else:
                    match = hdr_val.search(line.decode('latin-1'))
                    if len(match.groups()) < 2:
                        hdr[match[1]] = None
                    else:
                        hdr[match[1]] = match[2]
        if verbose:
            print(hdr)

        data = np.fromfile(fp)

    data_cmplx = data[0::2] + 1j * data[1::2]

    # MRSI
    if (int(hdr['CSIMatrixSize[0]'])
            * int(hdr['CSIMatrixSize[1]'])
            * int(hdr['CSIMatrixSize[2]']))\
            > 1:
        data_cmplx = data_cmplx.reshape((
            int(hdr['CSIMatrixSize[2]']),
            int(hdr['CSIMatrixSize[1]']),
            int(hdr['CSIMatrixSize[0]']),
            int(hdr['VectorSize'])))
        data_cmplx = np.moveaxis(data_cmplx, (0, 1, 2), (2, 1, 0))
        data_shape = data_cmplx.shape[:3]

        imagePositionPatient = np.asarray([
            _locale_float(hdr['PositionVector[0]']),
            _locale_float(hdr['PositionVector[1]']),
            _locale_float(hdr['PositionVector[2]'])])

        half_shift = True
    # SVS
    else:
        data_cmplx = data_cmplx.reshape((1, 1, 1) + data_cmplx.shape)
        data_shape = (1, 1, 1)

        try:
            imagePositionPatient = np.asarray([
                _locale_float(hdr['VOIPositionSag']),
                _locale_float(hdr['VOIPositionCor']),
                _locale_float(hdr['VOIPositionTra'])])
        except KeyError:
            imagePositionPatient = np.asarray([
                _locale_float(hdr['PositionVector[0]']),
                _locale_float(hdr['PositionVector[1]']),
                _locale_float(hdr['PositionVector[2]'])])

        half_shift = False

    dwelltime = _locale_float(hdr['DwellTime']) / 1E6

    imageOrientationPatient = np.asarray([
        [_locale_float(hdr['RowVector[0]']), _locale_float(hdr['ColumnVector[0]'])],
        [_locale_float(hdr['RowVector[1]']), _locale_float(hdr['ColumnVector[1]'])],
        [_locale_float(hdr['RowVector[2]']), _locale_float(hdr['ColumnVector[2]'])]]).T

    xyzMM = np.asarray([
        _locale_float(hdr['PixelSpacingRow']),
        _locale_float(hdr['PixelSpacingCol']),
        _locale_float(hdr['SliceThickness'])])

    currNiftiOrientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                                    imagePositionPatient,
                                                    xyzMM,
                                                    data_shape,
                                                    verbose=verbose,
                                                    half_shift=half_shift)

    meta = extractRdaMetadata(hdr)
    meta.set_standard_def('OriginalFile', [rda_path.name])

    if fname_out:
        name = fname_out
    else:
        name = rda_path.stem

    return [gen_nifti_mrs_hdr_ext(data_cmplx, dwelltime, meta, currNiftiOrientation.Q44, no_conj=True), ], [name, ]


def extractRdaMetadata(hdr):
    """Extract information from the rda header dict to insert into the json header ext.

    :param hdr: Header read from rda file
    :type hdr: dict
    :return: Complete nifti_mrs header object
    :rtype: nifti_mrs.nifti_mrs.Hdr_Ext
    """

    # Extract required metadata and create hdr_ext object
    obj = Hdr_Ext(
        _locale_float(hdr['MRFrequency']),
        hdr['Nucleus'])

    # Standard defined metadata
    def set_standard_def(nifti_mrs_key, key, cast=None):
        if key in hdr\
                and hdr[key] is not None\
                and hdr[key]:
            if cast is not None:
                obj.set_standard_def(nifti_mrs_key, cast(hdr[key]))
            else:
                obj.set_standard_def(nifti_mrs_key, hdr[key])

    # # 5.1 MRS specific Tags
    # 'EchoTime'
    obj.set_standard_def('EchoTime', _locale_float(hdr['TE']) * 1E-3)
    # 'RepetitionTime'
    obj.set_standard_def('RepetitionTime', _locale_float(hdr['TR']) * 1E-3)
    # 'InversionTime'
    obj.set_standard_def('InversionTime', _locale_float(hdr['TI']) * 1E-3)
    # 'MixingTime'
    obj.set_standard_def('MixingTime', _locale_float(hdr['TM']) * 1E-3)
    # 'ExcitationFlipAngle'
    obj.set_standard_def('ExcitationFlipAngle', _locale_float(hdr['FlipAngle']))
    # 'TxOffset'
    # 'VOI'
    # 'WaterSuppressed'
    # 'WaterSuppressionType'
    # 'SequenceTriggered'

    # # 5.2 Scanner information
    # 'Manufacturer'
    obj.set_standard_def('Manufacturer', 'Siemens')
    # 'ManufacturersModelName'
    obj.set_standard_def('ManufacturersModelName', hdr['ModelName'])
    # 'DeviceSerialNumber'
    obj.set_standard_def('DeviceSerialNumber', hdr['DeviceSerialNumber'])
    # 'SoftwareVersions'
    obj.set_standard_def('SoftwareVersions', hdr['SoftwareVersion[0]'])
    # 'InstitutionName'
    obj.set_standard_def('InstitutionName', hdr['InstitutionName'])
    # 'InstitutionAddress'
    # 'TxCoil'
    obj.set_standard_def('TxCoil', hdr['TransmitCoil'])
    # 'RxCoil'

    # # 5.3 Sequence information
    # 'SequenceName'
    set_standard_def('SequenceName', 'SequenceName')
    # 'ProtocolName'
    set_standard_def('ProtocolName', 'ProtocolName')
    # # 5.4 Sequence information
    # 'PatientPosition'
    set_standard_def('PatientPosition', 'PatientPosition')
    # 'PatientName'
    set_standard_def('PatientName', 'PatientName')
    # 'PatientID'
    set_standard_def('PatientID', 'PatientID')
    # 'PatientWeight'
    set_standard_def('PatientWeight', 'PatientWeight', cast=_locale_float)
    # 'PatientDoB'
    set_standard_def('PatientDoB', 'PatientBirthDate')
    # 'PatientSex'
    set_standard_def('PatientSex', 'PatientSex')

    # # 5.5 Provenance and conversion metadata
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)

    # 'OriginalFile'
    # Set elsewhere

    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    return obj



'''Tests for Siemens twix format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
siemens_path = Path(__file__).parent / 'spec2nii_test_data' / 'Siemens'
vb_path = siemens_path / 'VBData' / 'Twix/meas_MID151_svs_se_C_T15_S10_10_FID108741.dat'
ve_path = siemens_path / 'VEData' / 'Twix/meas_MID00240_FID62745_svs_se_c_t15_s10_R10.dat'
xa20_path = siemens_path / 'XAData' / 'XA20/twix/meas_MID00027_FID07653_svs_se_30_WS_on.dat'
xa30_path = siemens_path / 'XAData' / 'XA30/meas_MID00479_FID106847_svs_se_135sws.dat'
ve_fid = siemens_path / 'fid' / 'meas_MID00070_FID27084_fid_13C_360dyn_hyper_TR1000.dat'

# Special cased data
hercules_ve = siemens_path / 'HERCULES' / 'Siemens_TIEMO_HERC.dat'
hercules_xa30 = siemens_path / 'HERCULES' / 'meas_MID02595_FID60346_HERC.dat'
hermes_xa50 = siemens_path / 'XAData' / 'XA50' / 'smm_svs_herc_v2_hermes.dat'


def test_VB(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'vb',
                           '-o', tmp_path,
                           '-j', str(vb_path)])

    img_t = read_nifti_mrs(tmp_path / 'vb.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 1056, 32, 2)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'


def test_VE(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 've',
                           '-o', tmp_path,
                           '-j', str(ve_path)])

    img_t = read_nifti_mrs(tmp_path / 've.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 1056, 32, 16)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'


def test_XA20(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'xa20',
                           '-o', tmp_path,
                           '-j', str(xa20_path)])

    img_t = read_nifti_mrs(tmp_path / 'xa20.nii.gz')
    img_ref = read_nifti_mrs(tmp_path / 'xa20_ref.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 2080, 42, 80)
    assert np.iscomplexobj(img_t.dataobj)
    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'

    hdr_ext_codes = img_ref.header.extensions.get_codes()
    hdr_ext = json.loads(img_ref.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_ref.shape == (1, 1, 1, 2080, 42)
    assert np.iscomplexobj(img_ref.dataobj)
    assert hdr_ext['dim_5'] == 'DIM_COIL'


def test_XA30(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'xa30',
                           '-o', tmp_path,
                           '-j', str(xa30_path)])

    img_t = read_nifti_mrs(tmp_path / 'xa30.nii.gz')
    img_ref = read_nifti_mrs(tmp_path / 'xa30_ref.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 4096, 44, 16)
    assert np.iscomplexobj(img_t.dataobj)
    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'

    hdr_ext_codes = img_ref.header.extensions.get_codes()
    hdr_ext = json.loads(img_ref.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_ref.shape == (1, 1, 1, 4096, 44)
    assert np.iscomplexobj(img_ref.dataobj)
    assert hdr_ext['dim_5'] == 'DIM_COIL'


def test_os_remove(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 've',
                           '-o', tmp_path,
                           '-j', str(ve_path),
                           '--remove_os'])

    img_t = read_nifti_mrs(tmp_path / 've.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 528, 32, 16)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'


def test_VE_fid(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 've_fid',
                           '-o', tmp_path,
                           '-j', str(ve_fid)])

    img_t = read_nifti_mrs(tmp_path / 've_fid.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 4096, 4, 10)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'


def test_VE_HERCULES(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'hercules_ve',
                           '-o', tmp_path,
                           '-j', str(hercules_ve)])

    img_t = read_nifti_mrs(tmp_path / 'hercules_ve.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 2080, 32, 4, 56)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_EDIT'
    assert hdr_ext['dim_7'] == 'DIM_DYN'


def test_XA_HERCULES(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'hercules_xa',
                           '-o', tmp_path,
                           '-j', str(hercules_xa30)])

    img_t = read_nifti_mrs(tmp_path / 'hercules_xa.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 2080, 32, 4, 56)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_EDIT'
    assert hdr_ext['dim_7'] == 'DIM_DYN'


def test_XA_HERMES(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'hermes_xa',
                           '-o', tmp_path,
                           '-j', str(hermes_xa50)])

    img_t = read_nifti_mrs(tmp_path / 'hermes_xa.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 4096, 42, 4, 80)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_EDIT'
    assert hdr_ext['dim_7'] == 'DIM_DYN'


def test_twix_mrsi_orientation(tmp_path):
    '''Test that the (empty) mrsi has information matching the twix equivalent.'''

    def run_comparison(file_twix, file_dcm):
        subprocess.run(
            ['spec2nii', 'dicom',
             '-o', tmp_path,
             '-f', 'dicom',
             file_dcm])

        subprocess.run(
            ['spec2nii', 'twix',
             '-e', 'image',
             '-o', tmp_path,
             '-f', 'twix',
             file_twix])

        img_d = read_nifti_mrs(tmp_path / 'dicom.nii.gz')
        img_t = read_nifti_mrs(tmp_path / 'twix_empty.nii.gz')

        assert np.allclose(img_t.affine, img_d.affine, atol=1E-3)
        assert img_t.shape == img_d.shape

    vb_pairs = [
        ['meas_MID145_csi_se_Tra_sat_FID108735.dat', 'csi_se_Tra_sat_9_1'],
        ['meas_MID143_csi_se_3D_C_S23_5_T20_3_10_FID108733.dat', 'csi_se_3D_C>S23.5>T20.3_10_8_1'],
        ['meas_MID141_csi_se_3D_S_T23_5_C20_3_10_FID108731.dat', 'csi_se_3D_S>T23.5>C20.3_10_7_1'],
        ['meas_MID139_csi_se_3D_T_C23_5_S20_3_10_FID108729.dat', 'csi_se_3D_T>C23.5>S20.3_10_6_1'],
        ['meas_MID137_csi_se_C_S23_5_T20_3_10_FID108727.dat', 'csi_se_C>S23.5>T20.3_10_5_1'],
        ['meas_MID135_csi_se_S_T23_5_C20_3_10_FID108725.dat', 'csi_se_S>T23.5>C20.3_10_4_1'],
        ['meas_MID133_csi_se_T_C23_5_S20_3_10_FID108723.dat', 'csi_se_T>C23.5>S20.3_10_3_1']]

    for pair in vb_pairs:
        twix = siemens_path / 'VBData' / 'Twix' / pair[0]
        dcm = siemens_path / 'VBData' / 'DICOM' / pair[1]
        run_comparison(twix, dcm)

    ve_pairs = [
        ['meas_MID00223_FID62728_csi_se_t_c23_5_s20_3_R10.dat', 'csi_se_t>c23.5>s20.3_R10_4_1'],
        ['meas_MID00225_FID62730_csi_se_s_t23_5_c20_3_R10.dat', 'csi_se_s>t23.5>c20.3_R10_5_1'],
        ['meas_MID00227_FID62732_csi_se_c_s23_5_t20_3_R10.dat', 'csi_se_c>s23.5>t20.3_R10_6_1'],
        ['meas_MID00229_FID62734_csi_se_3D_t_c23_5_s20_3_R10.dat', 'csi_se_3D_t>c23.5>s20.3_R10_7_1'],
        ['meas_MID00231_FID62736_csi_se_3D_s_t23_5_c20_3_R10.dat', 'csi_se_3D_s>t23.5>c20.3_R10_8_1'],
        ['meas_MID00233_FID62738_csi_se_3D_c_s23_5_t20_3_R10.dat', 'csi_se_3D_c>s23.5>t20.3_R10_9_1'],
        ['meas_MID00244_FID62749_csi_se_iso_tra_sat.dat', 'csi_se_iso_tra_sat_14_1']]

    for pair in ve_pairs:
        twix = siemens_path / 'VEData' / 'Twix' / pair[0]
        dcm = siemens_path / 'VEData' / 'DICOM' / pair[1]
        run_comparison(twix, dcm)



"""Run tests for UIH data.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
"""

from pathlib import Path
import subprocess
from .io_for_tests import read_nifti_mrs
from PIL import Image
import pytest

output_path = Path(__file__).parent / 'orientation_img'
output_path.mkdir(exist_ok=True)

data_base = Path(__file__).parent / 'spec2nii_test_data' / 'UIH'

data_paths = {'svs': data_base / 'mrs_data/dicom/svs_press_te144_SVS_801/00000001.dcm',
              'csi_2d': data_base / 'mrs_data/dicom/csi_hise_te144_CSI_1201/00000000.dcm',
              'csi_3d': data_base / 'mrs_3d/dicom/csi_hise_3d_te144_CSI_1301/00000000.dcm'}

screenshots = {'svs': data_base / 'mrs_data/svs_te144_Image_0.jpg',
               'csi_2d': data_base / 'mrs_data/csi_te144_Image_0.jpg',
               'csi_3d': data_base / 'mrs_3d/csi3d_te144_Image_0.jpg'}


def get_concat_v(im1, im2):
    dst = Image.new('RGB', (max(im1.width, im2.width), im1.height + im2.height))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (0, im1.height))
    return dst


def flip_first_third(img_in):
    width, height = img_in.size
    cropped_0 = img_in.crop((0, 0, int(width / 3), height))
    cropped_0 = cropped_0.transpose(Image.FLIP_LEFT_RIGHT)
    cropped_1 = img_in.crop((int(width / 3) + 1, 0, width, height))
    out = Image.new('RGB', (width, height))
    out.paste(cropped_0, (0, 0))
    out.paste(cropped_1, (int(width / 3) + 1, 0))
    return out


@pytest.mark.orientation
def test_uih_svs(tmp_path):

    # Convert the data
    subprocess.check_call(['spec2nii', 'uih',
                           '-f', 'svs',
                           '-o', tmp_path,
                           data_paths['svs']])

    # Check data size
    img = read_nifti_mrs(tmp_path / 'svs.nii.gz')
    assert img.shape == (1, 1, 1, 2048)

    # Manual orientation check
    # Make fsleyes rendering
    subprocess.check_call(['fsleyes', 'render',
                           '-of', tmp_path / 'svs.png',
                           '-vl', '245', '296', '10',
                           '-hc', data_base / 'mrs_data/2D_tra.nii.gz',
                           tmp_path / 'svs.nii.gz', '-ot', 'complex',
                           '-a', '50', '-cm', 'blues'])

    fsl_render = flip_first_third(Image.open(tmp_path / 'svs.png'))
    screenshot = Image.open(screenshots['svs'])
    screenshot = screenshot.crop((0, 0, screenshot.width, int(screenshot.height / 2)))
    combined = get_concat_v(fsl_render, screenshot)

    combined.save(output_path / 'uih_svs.png')


@pytest.mark.orientation
def test_uih_2d_csi(tmp_path):

    # Convert the data
    subprocess.check_call(['spec2nii', 'uih',
                           '-f', 'csi_2d',
                           '-o', tmp_path,
                           data_paths['csi_2d']])

    # Check data size
    img = read_nifti_mrs(tmp_path / 'csi_2d.nii.gz')
    assert img.shape == (16, 16, 1, 2048)

    # Manual orientation check
    # Make fsleyes rendering
    subprocess.check_call(['fsleyes', 'render',
                           '-of', tmp_path / 'csi_2d.png',
                           '-vl', '245', '296', '9',
                           '-hc', data_base / 'mrs_data/2D_tra.nii.gz',
                           tmp_path / 'csi_2d.nii.gz', '-ot', 'complex',
                           '-a', '50', '-cm', 'blue', '-dr', '-0.9', '9.7'])

    fsl_render = flip_first_third(Image.open(tmp_path / 'csi_2d.png'))
    screenshot = Image.open(screenshots['csi_2d'])
    screenshot = screenshot.crop((0, 0, screenshot.width, int(screenshot.height / 2)))
    combined = get_concat_v(fsl_render, screenshot)

    combined.save(output_path / 'uih_csi_2d.png')


@pytest.mark.orientation
def test_uih_3d_csi(tmp_path):

    # Convert the data
    subprocess.check_call(['spec2nii', 'uih',
                           '-f', 'csi_3d',
                           '-o', tmp_path,
                           data_paths['csi_3d']])

    # Check data size
    img = read_nifti_mrs(tmp_path / 'csi_3d.nii.gz')
    assert img.shape == (8, 8, 8, 1024)

    # Manual orientation check
    # Make fsleyes rendering
    subprocess.check_call(['fsleyes', 'render',
                           '-of', tmp_path / 'csi_3d.png',
                           '-vl', '245', '296', '9',
                           '-hc', data_base / 'mrs_3d/3d_tra.nii.gz',
                           tmp_path / 'csi_3d.nii.gz', '-ot', 'complex',
                           '-a', '50', '-cm', 'blue', '-dr', '-0.9', '9.7'])

    fsl_render = flip_first_third(Image.open(tmp_path / 'csi_3d.png'))
    screenshot = Image.open(screenshots['csi_3d'])
    screenshot = screenshot.crop((0, 0, screenshot.width, int(screenshot.height / 2)))
    combined = get_concat_v(fsl_render, screenshot)

    combined.save(output_path / 'uih_csi_3d.png')



"""Module containing other miscellaneous functions

Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2021 University of Oxford
"""
import json
import pprint

from nifti_mrs.nifti_mrs import NIFTI_MRS, NotNIFTI_MRS


def dump_headers(args):
    """Print the nifti headers and header extensions

    :param args: Command line arguments
    """
    # Load file
    nifti_mrs_img = NIFTI_MRS(args.file, validate_on_creation=False)

    # Print NIfTI header
    print('NIfTI header: ')
    print(nifti_mrs_img.header)

    print('\nNIfTI-MRS header extension: ')
    pp = pprint.PrettyPrinter()
    pp.pprint(nifti_mrs_img.hdr_ext.to_dict())


def extract_hdr_ext(args):
    """Generate a json sidecar file for input file

    :param args: Command line arguments
    """
    nifti_mrs_img = NIFTI_MRS(args.file, validate_on_creation=False)

    if args.outdir:
        args.outdir.mkdir(exist_ok=True, parents=True)
        out_json = args.outdir / (args.file.with_suffix('').with_suffix('').name + '.json')
    else:
        out_json = args.file.parent / (args.file.with_suffix('').with_suffix('').name + '.json')

    with open(out_json, 'w') as fp:
        json.dump(nifti_mrs_img.hdr_ext.to_dict(), fp, indent=4)


def insert_hdr_ext(args):
    """Function for inserting a new header into NIfTI-MRS files.

    :param args: Command line arguments parsed in spec2nii.py
    :return: Modified NIfTI-MRS file
    :rtype: [nifti_mrs.nifti_mrs.NIFTI_MRS,]
    :return: List of output names
    :rtype: [str,]
    """

    with open(args.json_file) as jf:
        new_hdr = json.load(jf)

    # Load data
    try:
        nifti_mrs_img = NIFTI_MRS(args.file, validate_on_creation=False)
        nifti_mrs_img.hdr_ext = new_hdr

    except NotNIFTI_MRS:
        print(f'{args.file} is not compliant with the NIfTI-MRS standard, attempting to convert')
        from fsl.data.image import Image
        from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
        from nifti_mrs.hdr_ext import Hdr_Ext

        nimg = Image(args.file)
        hdr_ext = Hdr_Ext.from_header_ext(new_hdr)
        nifti_mrs_img = gen_nifti_mrs_hdr_ext(
            nimg[:],
            nimg.header['pixdim'][4],
            hdr_ext,
            affine=nimg.getAffine('voxel', 'world'))

    if args.dwelltime is not None:
        nifti_mrs_img.dwelltime = args.dwelltime

    # Process output name.
    if args.fileout:
        fname_out = [args.fileout, ]
    else:
        fname_out = [args.file.with_suffix('').with_suffix('').name, ]

    return [nifti_mrs_img, ], fname_out



'''Tests for Siemens rda format conversion.

Copyright William Clarke, University of Oxford 2022
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
siemens_path = Path(__file__).parent / 'spec2nii_test_data' / 'Siemens'

xa20_svs_path = siemens_path / 'XAData' / 'XA20/rda/spct_002.MR.MRI-LAB Test_Dir.5.1.114540.rda'
xa31_locale_svs_path = siemens_path / 'XAData' / 'XA31/rda/locale_XA31.rda'
xa50_path = siemens_path / 'XAData' / 'XA50' / 'Phantom_20240129.MR.10.1.155227.rda'

latin1_encoding = siemens_path / 'rda' / 'latin1.rda'

ve_path = siemens_path / 'VEData' / 'rda'

paired_ve_data = {
    'svs_se_c_t15_s10_r10': 'svs_se_c>t15>s10_R10_12_1',
    'svs_se_iso_tra_sat': 'svs_se_iso_tra_sat_13_1',
    'svs_se_s_c10_t5_r10': 'svs_se_s>c10>t5_R10_11_1',
    'svs_se_t_c15_s10_r10': 'svs_se_t>c15>s10_R10_10_1',
    'csi_se_3d_c_s235_t203_r10': 'csi_se_3D_c>s23.5>t20.3_R10_9_1',
    'csi_se_3d_s_t235_c203_r10': 'csi_se_3D_s>t23.5>c20.3_R10_8_1',
    'csi_se_3d_t_c235_s203_r10': 'csi_se_3D_t>c23.5>s20.3_R10_7_1',
    'csi_se_c_s235_t203_r10': 'csi_se_c>s23.5>t20.3_R10_6_1',
    'csi_se_iso_tra_sat': 'csi_se_iso_tra_sat_14_1',
    'csi_se_s_t235_c203_r10': 'csi_se_s>t23.5>c20.3_R10_5_1',
    'csi_se_t_c235_s203_r10': 'csi_se_t>c23.5>s20.3_R10_4_1'}


def test_svs_ve_against_dicom(tmp_path):
    for rda_file in ve_path.glob("svs_se*.rda"):
        print(rda_file.stem)
        dcm_dir = rda_file.parent.parent / 'DICOM' / paired_ve_data[rda_file.stem]

        subprocess.run([
            'spec2nii', 'rda',
            '-f', rda_file.stem + '_rda',
            '-o', tmp_path,
            '-j', rda_file])

        subprocess.run([
            'spec2nii', 'dicom',
            '-f', rda_file.stem + '_dcm',
            '-o', tmp_path,
            '-j', dcm_dir])

        rda_out = tmp_path / (rda_file.stem + '_rda.nii.gz')
        dcm_out = tmp_path / (rda_file.stem + '_dcm.nii.gz')
        assert rda_out.exists()
        assert dcm_out.exists()

        dcm = read_nifti_mrs(dcm_out)
        rda = read_nifti_mrs(rda_out)
        assert dcm.shape == rda.shape
        assert np.allclose(dcm.get_fdata(dtype=complex), rda.get_fdata(dtype=complex))
        assert np.allclose(dcm.get_sform(), rda.get_sform())

        hdr_ext_codes = dcm.header.extensions.get_codes()
        hdr_ext_dcm = json.loads(dcm.header.extensions[hdr_ext_codes.index(44)].get_content())
        hdr_ext_codes = rda.header.extensions.get_codes()
        hdr_ext_rda = json.loads(rda.header.extensions[hdr_ext_codes.index(44)].get_content())

        for key in ['SpectrometerFrequency', 'ResonantNucleus', 'EchoTime', 'RepetitionTime']:
            assert hdr_ext_dcm[key] == hdr_ext_rda[key]


def test_csi_ve_against_dicom(tmp_path):
    for rda_file in ve_path.glob("csi_se*.rda"):
        print(rda_file.stem)
        dcm_dir = rda_file.parent.parent / 'DICOM' / paired_ve_data[rda_file.stem]

        subprocess.run([
            'spec2nii', 'rda',
            '-f', rda_file.stem + '_rda',
            '-o', tmp_path,
            '-j', rda_file])

        subprocess.run([
            'spec2nii', 'dicom',
            '-f', rda_file.stem + '_dcm',
            '-o', tmp_path,
            '-j', dcm_dir])

        rda_out = tmp_path / (rda_file.stem + '_rda.nii.gz')
        dcm_out = tmp_path / (rda_file.stem + '_dcm.nii.gz')
        assert rda_out.exists()
        assert dcm_out.exists()

        dcm = read_nifti_mrs(dcm_out)
        rda = read_nifti_mrs(rda_out)
        assert dcm.shape == rda.shape

        assert np.allclose(dcm.get_fdata(dtype=complex), rda.get_fdata(dtype=complex))
        assert np.allclose(dcm.get_sform(), rda.get_sform(), atol=1E-4)

        hdr_ext_codes = dcm.header.extensions.get_codes()
        hdr_ext_dcm = json.loads(dcm.header.extensions[hdr_ext_codes.index(44)].get_content())
        hdr_ext_codes = rda.header.extensions.get_codes()
        hdr_ext_rda = json.loads(rda.header.extensions[hdr_ext_codes.index(44)].get_content())

        for key in ['SpectrometerFrequency', 'ResonantNucleus', 'EchoTime', 'RepetitionTime']:
            assert hdr_ext_dcm[key] == hdr_ext_rda[key]


def test_xa20_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'rda',
                           '-f', 'xa_svs',
                           '-o', tmp_path,
                           '-j', str(xa20_svs_path)])

    img_t = read_nifti_mrs(tmp_path / 'xa_svs.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    hdr_ext['ResonantNucleus'] = ['1H', ]

    assert img_t.shape == (1, 1, 1, 1024)
    assert np.iscomplexobj(img_t.dataobj)


def test_xa31_locale_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'rda',
                           '-f', 'xa_svs',
                           '-o', tmp_path,
                           '-j', str(xa31_locale_svs_path)])

    img_t = read_nifti_mrs(tmp_path / 'xa_svs.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    hdr_ext['ResonantNucleus'] = ['1H', ]

    assert img_t.shape == (1, 1, 1, 1024)
    assert np.iscomplexobj(img_t.dataobj)


def test_xa50_svs(tmp_path):

    subprocess.run([
        'spec2nii', 'rda',
        '-f', 'xa50_svs',
        '-o', tmp_path,
        '-j', xa50_path])

    img_t = read_nifti_mrs(tmp_path / 'xa50_svs.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    hdr_ext['ResonantNucleus'] = ['1H', ]

    assert img_t.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img_t.dataobj)


def test_latin_encoding(tmp_path):

    subprocess.check_call(['spec2nii', 'rda',
                           '-f', 'rda_latin1',
                           '-o', tmp_path,
                           '-j', str(latin1_encoding)])

    img_t = read_nifti_mrs(tmp_path / 'rda_latin1.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    hdr_ext['ResonantNucleus'] = ['1H', ]

    assert img_t.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img_t.dataobj)



'''Tests for Siemens dicom format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
siemens_path = Path(__file__).parent / 'spec2nii_test_data' / 'Siemens'
vb_svs_path = siemens_path / 'VBData' / 'DICOM/svs_se_C>T15>S10_10_12_1/'
vb_mrsi_path = siemens_path / 'VBData' / 'DICOM/csi_se_3D_C>S23.5>T20.3_10_8_1/'

ve_svs_path = siemens_path / 'VEData' / 'DICOM/svs_se_s>c10>t5_R10_11_1/'
ve_mrsi_path = siemens_path / 'VEData' / 'DICOM/csi_se_3D_t>c23.5>s20.3_R10_7_1/'

xa20_svs_path = siemens_path / 'XAData' / 'XA20/DICOM/26516628.dcm'
xa30_svs_path = siemens_path / 'XAData' / 'XA30/meas_MID00479_FID106847_svs_se_135sws.dcm'

tim_trio_non_image = siemens_path / 'VBData' / 'tim_trio_sop_class_uid' / 'timtrio_mr_vb19a.dcm'
tim_trio_mrspecstorage = siemens_path / 'VBData' / 'tim_trio_sop_class_uid' / 'timtrio_mrdcm_vb13a_2_0.dcm'

enhanced_csi_1 = siemens_path / 'enhanced_dcm_csi' / 'rk_enhanced' / '1.dcm'
enhanced_csi_2 = siemens_path / 'enhanced_dcm_csi'\
    / 'sm_enhanced' / 'Phantom.Psychiatrie^Studien^aslac.Sr27.I1.20231006084036808.dcm'
enhanced_csi_2_comp = siemens_path / 'enhanced_dcm_csi'\
    / 'sm_classic' / 'PHANTOM.MR.PSYCHIATRIE_STUDIEN_ASLAC.0027.0001.2023.10.06.08.40.21.250955.407258759.IMA'


def test_VB_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'vb_svs',
                           '-o', tmp_path,
                           '-j', str(vb_svs_path)])

    img_t = read_nifti_mrs(tmp_path / 'vb_svs.nii.gz')

    # hdr_ext_codes = img_t.header.extensions.get_codes()
    # hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 512)
    assert np.iscomplexobj(img_t.dataobj)


def test_VE_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 've_svs',
                           '-o', tmp_path,
                           '-j', str(ve_svs_path)])

    img_t = read_nifti_mrs(tmp_path / 've_svs.nii.gz')

    # hdr_ext_codes = img_t.header.extensions.get_codes()
    # hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 512)
    assert np.iscomplexobj(img_t.dataobj)


def test_XA20_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'xa20_svs',
                           '-o', tmp_path,
                           '-j', str(xa20_svs_path)])

    img_t = read_nifti_mrs(tmp_path / 'xa20_svs.nii.gz')

    assert img_t.shape == (1, 1, 1, 1024)
    assert np.iscomplexobj(img_t.dataobj)


def test_XA30_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'xa30_svs',
                           '-o', tmp_path,
                           '-j', str(xa30_svs_path)])

    img_t = read_nifti_mrs(tmp_path / 'xa30_svs.nii.gz')

    assert img_t.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img_t.dataobj)


def test_VB_mrsi(tmp_path):

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'vb_mrsi',
                           '-o', tmp_path,
                           '-j', str(vb_mrsi_path)])

    img_t = read_nifti_mrs(tmp_path / 'vb_mrsi.nii.gz')

    # hdr_ext_codes = img_t.header.extensions.get_codes()
    # hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (16, 16, 8, 512)
    assert np.iscomplexobj(img_t.dataobj)


def test_VE_mrsi(tmp_path):

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 've_mrsi',
                           '-o', tmp_path,
                           '-j', str(ve_mrsi_path)])

    img_t = read_nifti_mrs(tmp_path / 've_mrsi.nii.gz')

    # hdr_ext_codes = img_t.header.extensions.get_codes()
    # hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (16, 16, 16, 512)
    assert np.iscomplexobj(img_t.dataobj)


mrsi_voi_path = siemens_path / 'voi_in_mrsi' /\
    'F3T_2021_PH_016.MR.FMRIB_DEVELOPER_WILL.0004.0001.2021.07.01.16.43.12.374485.667204044.IMA'


def test_voi_generation(tmp_path):

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'mrsi',
                           '--voi',
                           '-o', tmp_path,
                           '-j', str(mrsi_voi_path)])

    img_t = read_nifti_mrs(tmp_path / 'mrsi.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert 'VOI' in hdr_ext
    assert np.asarray(hdr_ext['VOI']).shape == (4, 4)

    assert (tmp_path / 'mrsi_voi.nii.gz').exists()


def test_sop_class_vb(tmp_path):
    """Test the two different SOPClassUIDs arising from the same type of scanner.
    """
    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'sop_non_image',
                           '-o', tmp_path,
                           '-j', str(tim_trio_non_image)])

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'sop_mrspecstorage',
                           '-o', tmp_path,
                           '-j', str(tim_trio_mrspecstorage)])

    img_1 = read_nifti_mrs(tmp_path / 'sop_non_image.nii.gz')
    img_2 = read_nifti_mrs(tmp_path / 'sop_mrspecstorage.nii.gz')

    assert img_1.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img_1.dataobj)

    assert img_2.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img_2.dataobj)


def test_enhanced_csi(tmp_path):

    # First test look for right shape (16, 16, 1, 1024)
    subprocess.run(['spec2nii', 'dicom',
                    '-f', 'csi_1',
                    '-o', tmp_path,
                    '-j', enhanced_csi_1])

    img_1 = read_nifti_mrs(tmp_path / 'csi_1.nii.gz')
    assert img_1.shape == (16, 16, 1, 1024)
    assert np.iscomplexobj(img_1.dataobj)

    # Second test: compare with classic dicom export
    subprocess.run(['spec2nii', 'dicom',
                    '-f', 'csi_2',
                    '-o', tmp_path,
                    '-j', enhanced_csi_2])
    subprocess.run(['spec2nii', 'dicom',
                    '-f', 'csi_2_comp',
                    '-o', tmp_path,
                    '-j', enhanced_csi_2_comp])

    img_2 = read_nifti_mrs(tmp_path / 'csi_2.nii.gz')
    img_2_comp = read_nifti_mrs(tmp_path / 'csi_2_comp.nii.gz')

    hdr_ext_codes = img_2.header.extensions.get_codes()
    hdr_ext = json.loads(img_2.header.extensions[hdr_ext_codes.index(44)].get_content())
    hdr_ext_codes = img_2_comp.header.extensions.get_codes()
    hdr_ext_comp = json.loads(img_2_comp.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_2.shape == img_2_comp.shape
    assert np.allclose(img_2.dataobj[:], img_2_comp.dataobj[:])
    for key in hdr_ext:
        print(key)
        # Skip keys with known differences.
        if key in ('RxCoil', 'SequenceName', 'PatientID', 'ConversionTime', 'OriginalFile'):
            continue
        if isinstance(hdr_ext[key], (float, complex)):
            assert np.isclose(hdr_ext[key], hdr_ext_comp[key])
        else:
            assert hdr_ext[key] == hdr_ext_comp[key]



from . import _version
from .due import due, Doi

__version__ = _version.get_versions()['version']

# Register the duecredit citation for spec2nii
due.cite(Doi('10.1002/mrm.29418'), description='Multi-format in vivo MR spectroscopy conversion to NIFTI',
         path='spec2nii', version=__version__, tags=['reference-implementation'], cite_module=True)



'''Orientation tests for Philips DICOM and spar/sdat.
This runs conversion on both spar/sdat pairs and the equivalent dicom file.
Data is acquired on a phantom with structure.
Test data acquired by Georg Oeltzschner <goeltzs1@jhmi.edu>.

Copyright William Clarke, University of Oxford, 2023.
Subject to the BSD 3-Clause License.
'''

from pathlib import Path
from subprocess import run

import pytest
from PIL import Image, ImageDraw

# Data paths
output_path = Path(__file__).parent / 'orientation_img'
output_path.mkdir(exist_ok=True)

philips_path = Path(__file__).parent / 'spec2nii_test_data' / 'philips'
base_dir = philips_path / 'spar_dcm_orientation_tests'
ssdir = base_dir / 'screenshots'

data = {
    '0-0-0': {
        'spar': 'dbiex_40_2_raw_act.spar',
        'dcm_old': '4002-iso_50-80-30_rot-0-0-0/'
                   'MRs.1.3.46.670589.11.24058.5.22.3.1.17592.'
                   '20230724182705801384c0baf12-9ba8-4b65-8382-52ecbde14039.dcm',
        'dcm_classic': 'classic/XX_0175',
        'dcm_enhanced': 'enhanced/XX_0002',
        'slices': [75, 58, 110]},
    '30-0-0': {
        'spar': 'dbiex_41_2_raw_act.spar',
        'dcm_old': '4102-iso_50-80-30_rot-30-0-0/MRs.1.3.46.670589.11.24058.5.22.3.1.17592.2023072418275589143.dcm',
        'dcm_classic': 'classic/XX_0177',
        'dcm_enhanced': 'enhanced/XX_0004',
        'slices': [75, 54, 111]},
    '30-40-0': {
        'spar': 'dbiex_42_2_raw_act.spar',
        'dcm_old': '4202-iso_50-80-30_rot-30-40-0/MRs.1.3.46.670589.11.24058.5.22.3.1.17592.2023072418281299148.dcm',
        'dcm_classic': 'classic/XX_0179',
        'dcm_enhanced': 'enhanced/XX_0006',
        'slices': [75, 55, 114]},
    '30-40-20': {
        'spar': 'dbiex_43_2_raw_act.spar',
        'dcm_old': '4302-iso_50-80-30_rot-30-40-20/MRs.1.3.46.670589.11.24058.5.22.3.1.17592.2023072418283012153.dcm',
        'dcm_classic': 'classic/XX_0181',
        'dcm_enhanced': 'enhanced/XX_0008',
        'slices': [75, 54, 114]},
    '0-40-20': {
        'spar': 'dbiex_44_2_raw_act.spar',
        'dcm_old': '4402-iso_50-80-30_rot-0-40-20/MRs.1.3.46.670589.11.24058.5.22.3.1.17592.2023072418284925158.dcm',
        'dcm_classic': 'classic/XX_0183',
        'dcm_enhanced': 'enhanced/XX_0010',
        'slices': [75, 30, 114]},
    '30-0-20': {
        'spar': 'dbiex_45_2_raw_act.spar',
        'dcm_old': '4502-iso_50-80-30_rot-30-0-20/MRs.1.3.46.670589.11.24058.5.22.3.1.17592.2023072418290724163.dcm',
        'dcm_classic': 'classic/XX_0185',
        'dcm_enhanced': 'enhanced/XX_0012',
        'slices': [75, 31, 114]},
    '10-10-44': {
        'spar': 'dbiex_46_2_raw_act.spar',
        'dcm_old': '4602-iso_50-80-30_rot-10-10-44/'
                   'MRs.1.3.46.670589.11.24058.5.22.3.1.17592.'
                   '20230724182924311689fcc2be8-08ee-4f06-9b99-189fbc586a2a.dcm',
        'dcm_classic': 'classic/XX_0187',
        'dcm_enhanced': 'enhanced/XX_0014',
        'slices': [75, 31, 114]},
    '10-44-10': {
        'spar': 'dbiex_47_2_raw_act.spar',
        'dcm_old': '4702-iso_50-80-30_rot-10-44-10/'
                   'MRs.1.3.46.670589.11.24058.5.22.3.1.17592.'
                   '20230724182939481730a508d00-9382-4749-b98c-7e372f5ce8dc.dcm',
        'dcm_classic': 'classic/XX_0189',
        'dcm_enhanced': 'enhanced/XX_0016',
        'slices': [75, 54, 114]},
    '44-10-10': {
        'spar': 'dbiex_48_2_raw_act.spar',
        'dcm_old': '4802-iso_50-80-30_rot-44-10-10/'
                   'MRs.1.3.46.670589.11.24058.5.22.3.1.17592.'
                   '20230724182956501788cc0636e-c425-484d-859c-2ed8881598d6.dcm',
        'dcm_classic': 'classic/XX_0191',
        'dcm_enhanced': 'enhanced/XX_0018',
        'slices': [75, 54, 114]}
}


def run_render(mode, rot, position, working_dir):
    if mode == 'cor':
        out_str = '_cor.png'
        img = '_cor_20230724170625_202.nii'
        positions = ['500', str(position), '500']
        hide1 = '--hidex'
        hide2 = '--hidez'
    elif mode == 'tra':
        out_str = '_tra.png'
        img = '_tra_20230724170625_203.nii'
        positions = ['507', '616', str(position)]
        hide1 = '--hidex'
        hide2 = '--hidey'
    elif mode == 'sag':
        out_str = '_sag.png'
        img = '_sag_20230724170625_204.nii'
        positions = [str(position), '278', '255']
        hide1 = '--hidez'
        hide2 = '--hidey'

    run([
        'fsleyes', 'render',
        '-of', working_dir / (rot + out_str),
        '-vl', positions[0], positions[1], positions[2],
        '-hc', hide1, hide2,
        '-ds', 'world',
        base_dir / img,
        working_dir / f'dcm_e_{rot}.nii.gz',
        '-a', '30', '-cm', 'red',
        working_dir / f'dcm_c_{rot}.nii.gz',
        '-a', '30', '-cm', 'green',
        working_dir / f'spar_{rot}.nii.gz',
        '-a', '30', '-cm', 'blue'])


def run_fsleyes(case, working_dir):
    run_render('cor', case, data[case]['slices'][0], working_dir)
    run_render('tra', case, data[case]['slices'][1], working_dir)
    run_render('sag', case, data[case]['slices'][2], working_dir)

    cor = Image.open(working_dir / (case + '_cor.png')).crop((140, 40, 660, 560))
    tra = Image.open(working_dir / (case + '_tra.png')).crop((140, 40, 660, 560))
    sag = Image.open(working_dir / (case + '_sag.png')).crop((140, 40, 660, 560))

    combined_img = Image.new(
        'RGB',
        (cor.width + tra.width + sag.width, cor.height))
    combined_img.paste(cor, (0, 0))
    combined_img.paste(tra, (cor.width, 0))
    combined_img.paste(sag.transpose(Image.FLIP_LEFT_RIGHT), (cor.width + tra.width, 0))
    return combined_img


def load_crop_scannerss(case):
    ss = Image.open(ssdir / f'{case.replace("-", "_")}.jpg')
    width, height = ss.size
    return ss.crop((0.210 * width, 0.047 * height, width, 0.5319 * height))


def run_case(case, working_dir):
    img_fsleyes = run_fsleyes(case, working_dir)
    img_scanner = load_crop_scannerss(case)
    scaled_fsleyes = img_fsleyes.resize((
        img_scanner.width,
        int(img_fsleyes.height * (img_scanner.width / img_fsleyes.width))))
    case_img = Image.new(
        'RGB',
        (scaled_fsleyes.width, scaled_fsleyes.height + img_scanner.height))
    case_img.paste(img_scanner, (0, 0))
    case_img.paste(scaled_fsleyes, (0, img_scanner.height))
    return case_img


@pytest.mark.orientation
def test_dcm_spar_orientations(tmp_path):
    def dcm_call(x, name):
        dcm = base_dir / x
        run(
            ['spec2nii', 'philips_dcm',
             '-o', tmp_path,
             '-f', name,
             dcm]
        )

    def ss_call(x, name):
        spar = base_dir / x
        sdat = spar.with_suffix('.sdat')
        run(
            ['spec2nii', 'philips',
             '-o', tmp_path,
             '-f', name,
             sdat, spar]
        )

    for key in data:
        dcm_call(data[key]['dcm_enhanced'], f'dcm_e_{key}')
        dcm_call(data[key]['dcm_classic'], f'dcm_c_{key}')
        ss_call(data[key]['spar'], f'spar_{key}')

    all_cases = []
    for case in data:
        all_cases.append(run_case(case, tmp_path))

    final_img = Image.new('RGB', (all_cases[0].width * 2, all_cases[0].height * 4))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(all_cases):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (all_cases[0].width * c, all_cases[0].height * r))
        draw.text((10 + all_cases[0].width * c, 40 + all_cases[0].height * r), list(data.keys())[idx], (255, 0, 0))

    final_img.save(output_path / 'philips_dcm_spar_comparison.png')



"""spec2nii module containing NIFTI orientation calculations.

This is all derived from Chris Rorden's dcm2niix code.
He has been very generous creating and releasing this code in an open way.

Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""
import numpy as np
from spec2nii.nifti_orientation import NIFTIOrient


def dcm_to_nifti_orientation(imageOrientationPatient, imagePositionPatient, xyzMM, data_shape,
                             half_shift=False, verbose=False):
    """Convert DCM orientation parameters to nifti orientations.

    :arg numpy.ndarray imageOrientationPatient: DICOM imageOrientationPatient tag. Shape = 2,3.
    :arg numpy.ndarray imagePositionPatient: DICOM imagePositionPatient tag.
    :arg numpy.ndarray xyzMM: Array containing the voxel sizes as contained in the
    nibabel.nicom.dicomwrappers.Wrapper voxel_sizes property.
    :arg data_shape: Shape of the spatial dimensions (dimensions 0-2).
    :arg bool half_shift: Apply half voxel shift to the x and y dimension.
    :arg bool verbose: Print debugging output.

    :return: NIFTIOrient object.
    :rtype: NIFTIOrient
    """
    # in style of dcm2niix
    # 1) calculate Q44
    Q44 = nifti_dicom2mat(imageOrientationPatient, imagePositionPatient, xyzMM, verbose=verbose)

    # 2) calculate nifti quaternion parameters
    # From github.com/rordenlab/dcm2niix/blob/
    # 081c6300d0cf47088f0873cd586c9745498f637a/console/nii_dicom.cpp#L604
    _, Q44 = verify_slice_dir(Q44, data_shape, imagePositionPatient, verbose=verbose)
    Q44[:2, :] *= -1

    # 3) If required apply the half-voxel shift in the first two dimensions
    if half_shift:
        Q44 = apply_half_voxel_shift(Q44)

    if verbose:
        print(f'Final Q44:\n {Q44}')

    # 4) place in data class for nifti orientation parameters
    return NIFTIOrient(Q44)


def apply_half_voxel_shift(Q44):
    # v = np.asarray([-0.5, -0.5, 0]) @ Q44[:3, :3]
    # Q44[:3, 3] += v
    # v = np.asarray([0.5, 0.5, 0]) @ Q44[:3, :3]
    # Q44[:3, 3] += v
    v = np.asarray([0.5, 0.5, 0]) @ Q44[:3, :3].T
    Q44[:3, 3] += v
    return Q44


def nifti_dicom2mat(orient, patientPosition, xyzMM, verbose=False):
    """DICOM values to 4x4 nifti transformation matrix.
    As per
    https://github.com/rordenlab/dcm2niix/blob/7ce33ca5fa3bb2dd4e5410bd97bcf515c9e462d9/console/nifti1_io_core.cpp"""

    Q = np.zeros((3, 3))

    # Q.m[0][0] = orient[1]; Q.m[0][1] = orient[2] ; Q.m[0][2] = orient[3] ; // load Q
    # Q.m[1][0] = orient[4]; Q.m[1][1] = orient[5] ; Q.m[1][2] = orient[6];
    Q[:2, :] = orient  # Fill first two rows

    # normalize rows
    normQ = np.linalg.norm(Q, ord=2, axis=1, keepdims=True)
    normQ[normQ == 0.0] = 1.0
    Q /= normQ
    if verbose:
        np.set_printoptions(precision=2)
        print(f'Normalised ImageOrientationPatient in Q:\n {Q}')

    # row 3 is the cross product of rows 1 and 2
    Q[2, :] = np.cross(Q[0, :], Q[1, :])
    if verbose:
        print(f'After slice normal calculation. Q:\n {Q}')

    Q = Q.T
    if np.linalg.det(Q) < 0.0:
        Q[:, 2] = Q[:, 2] * -1

    # next scale matrix
    # I think dcm2niix reverses the pixel spacing
    # https://github.com/rordenlab/dcm2niix/blob/485c387c93bbca3b29b93403dfde211c4bc39af6/console/nii_dicom.cpp#L5403
    xyzMM[1], xyzMM[0] = xyzMM[0], xyzMM[1]
    diagVox = np.diag(xyzMM)
    Q = Q @ diagVox
    if verbose:
        print(f'After scaling. Q:\n {Q}')

    # Include translations
    Q44 = np.zeros((4, 4))
    Q44[:3, :3] = Q
    Q44[:3, 3] = patientPosition
    Q44[3, 3] = 1.0
    if verbose:
        print(f'Final Q:\n {Q44}')
    return Q44


def verify_slice_dir(R, dim, patPos, verbose=False):
    # returns slice direction: 0=sag,1=coronal,2=axial, -= flipped
    if dim[2] < 2:
        return None, R  # don't care about direction for single slice

    iSL = 0  # find Z-slice direction: row with highest magnitude of 3rd column
    if (np.abs(R[1, 2]) >= np.abs(R[0, 2]))\
            and (np.abs(R[1, 2]) >= np.abs(R[2, 2])):
        iSL = 1
    if (np.abs(R[2, 2]) >= np.abs(R[0, 2]))\
            and (np.abs(R[2, 2]) >= np.abs(R[1, 2])):
        iSL = 2

    if verbose:
        print(f'z-dir = {iSL}')

    pos = patPos[iSL]
    if verbose:
        print(f'pos = {pos}')

    x = np.array([0.0, 0.0, dim[2] - 1.0, 1.0])
    # breakpoint()
    pos1v = x @ R.T
    pos1 = pos1v[iSL]
    if verbose:
        print(f'pos1v:\n{pos1v}')
        print(f'pos1 = {pos1}')

    # flip = (pos > R[iSL, 3]) != (pos1 > R[iSL, 3])
    flip = pos1 < R[iSL, 3]
    if verbose:
        print(f'{pos} > {R[iSL, 3]}')
        print(f'{pos1} > {R[iSL, 3]}')

    if flip:
        R[:, 2] *= -1
        iSL *= -1

    if verbose:
        print(f'verify slice dir {dim}')
        print(f'flip = {flip}')
        print(f'sliceDir = {iSL}')
        print(f'pos1 = {pos1}')

    return iSL, R


def nii_flipY(img, Q44, dim):
    """Apply a flip in y to match the "pure conversion" like dcm2niix.
    Viewing the image should be identical as both the image and the affine
    matrix are flipped.
    """

    s = Q44[:3, :3]
    v = np.array([0.0, dim[1] - 1.0, 0.0, 1.0])

    v = v @ Q44
    mFlipY = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 1]])

    s = s @ mFlipY

    Q44[:3, :3] = s
    Q44[:3, 3] += v[:3]

    img = np.flip(img, axis=1)
    return img, Q44
